import inspect
import re
from aiohttp import ClientResponseError
from typing import Optional


_docre = re.compile(r"(?i)([\d]+):\s+([\w\s]+)\s+status\s+code")


def _find(cls, test):
    for clz in filter(test, cls.__subclasses__()):
        if clz.__subclasses__():
            return _find(clz, test)
        return clz
    return cls


class HTTPException(ClientResponseError):
    """
    Base exception for failed HTTP requests.
    """

    code = None
    codes = ()

    def __new__(cls, e: Optional[ClientResponseError] = None):
        if not e:
            return super().__new__(cls, e)
        if not isinstance(e, ClientResponseError):
            raise TypeError(type(e))

        def test(clz):
            if clz.code:
                return e.status == clz.code
            return e.status in clz.codes

        return super().__new__(_find(cls, test), e)

    def __init__(self, e: Optional[ClientResponseError] = None):
        d = {
            k: None if v.default == inspect.Parameter.empty else v.default
            for k, v in inspect.signature(ClientResponseError).parameters.items()
        }
        if e:
            d.update(e.__dict__)
        else:
            match = _docre.search(self.__class__.__doc__)
            if not match:
                raise TypeError("This class cannot be directly instantiated.")
            d.update({"status": int(match[1]), "message": match[2]})
        super().__init__(**d)


class ClientError(HTTPException):
    """
    :exc:`HTTPException` for 4XX: Client Error status codes.
    """

    codes = range(400, 500)


class BadRequest(ClientError):
    """
    :exc:`ClientError` for 400: Bad Request status codes.
    """

    code = 400


class Forbidden(ClientError):
    """
    :exc:`ClientError` for 403: Forbidden status codes.
    """

    code = 403


class NotFound(ClientError):
    """
    :exc:`ClientError` for 404: Not Found status codes.
    """

    code = 404


class Conflict(ClientError):
    """
    :exc:`ClientError` for 409: Conflict status codes.
    """

    code = 409


class Teapot(ClientError):
    """
    :exc:`ClientError` for 418: I'm a Teapot status codes.
    """

    code = 418


class TooManyRequests(ClientError):
    """
    :exc:`ClientError` for 429: Too Many Requests status codes.
    """

    code = 429


class ServerError(HTTPException):
    """
    :exc:`HTTPException` for 5XX: Server Error status codes.
    """

    codes = range(500, 600)



from distutils.version import StrictVersion as _ST


__title__ = "sans"
__author__ = "Zephyrkul"
__license__ = "MIT"
__copyright__ = "Copyright 2019-2019 Zephyrkul"

version_info = _ST("0.0.1b7")
__version__ = str(version_info)

API_URL = ("https", "www.nationstates.net", "/cgi-bin/api.cgi")
DEFAULT_LOCAL_URL = ("http", "localhost:5557", "/")



import asyncio
import contextlib
from lxml import etree
from threading import Thread
from typing import Optional


def pretty_string(element_or_tree: etree.ElementBase) -> str:
    """
    Returns the base XML as a formatted and indented string.
    """
    return etree.tostring(element_or_tree, encoding=str, pretty_print=True)


_get_running_loop = getattr(
    asyncio,
    "get_running_loop",
    getattr(asyncio, "_get_running_loop", asyncio.get_event_loop),
)


def get_running_loop() -> Optional[asyncio.AbstractEventLoop]:
    """
    Gets the running event loop without creating a new one.

    Backport of :func:`asyncio.get_running_loop` which doesn't raise.
    """
    with contextlib.suppress(RuntimeError):
        return _get_running_loop()
    return None


def run_in_thread() -> None:
    """
    Runs the API event loop in its own thread.

    Required in order to make synchronous requests.
    """
    from .api import Api

    if Api._loop:
        raise RuntimeError("API event loop is already set.")

    def run(loop):
        try:
            loop.run_forever()
        finally:
            asyncio.gather(
                *asyncio.all_tasks(loop=loop), loop=loop, return_exceptions=True
            ).cancel()
            loop.run_until_complete(Api.session.close())
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    loop = asyncio.new_event_loop()
    Api.loop = loop
    Thread(target=run, args=(loop,), name="NS API Event Loop", daemon=True).start()



import argparse
import logging
import shlex
import sys

import sans
from sans.api import Api
from sans.errors import HTTPException
from sans.utils import pretty_string


logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="SANS Console Entry",
        prog="sans",
        epilog="Any unknown args will be used to build the API request.",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + sans.__version__
    )
    parser.add_argument("--agent", help="set the script's user agent")
    parser.add_argument(
        "--quit",
        action="store_true",
        help="quit the console program loop after this run",
    )
    parser.usage = parser.format_usage() + " ..."
    known, unknown = parser.parse_known_args()
    if not any(vars(known).values()):
        parser.print_help(sys.stderr)
    sans.run_in_thread()
    if known.quit and not unknown:
        return print("Okay I'll just leave I guess...", file=sys.stderr)
    Api.agent = known.agent or input("User agent: ")
    args = ()
    while True:
        print()
        if not args and unknown:
            args = (known, unknown)
        else:
            args = parser.parse_known_args(shlex.split(input(parser.prog + " ")))
            if args[0].agent:
                print(
                    "You can't change the agent in the middle of the script.",
                    file=sys.stderr,
                )
        known, unknown = args
        if not unknown:
            if known.quit:
                return print("No query provided. Exiting...", file=sys.stderr)
            print("No query provided.", file=sys.stderr)
            parser.print_usage(sys.stderr)
            continue
        request = {}
        key = "q"
        for arg in unknown:
            if arg.startswith("--"):
                if key != "q":
                    print("No value provided for key {!r}".format(key), file=sys.stderr)
                    continue
                key = arg[2:]
            else:
                request.setdefault(key, []).append(arg)
                key = "q"
        if key != "q":
            print("No value provided for key {!r}".format(key), file=sys.stderr)
            continue
        request = Api(request)
        try:
            for element in request.threadsafe:
                print(pretty_string(element), end="")
        except HTTPException as e:
            print(e, file=sys.stderr)
        except Exception as e:
            logging.exception(e)
        if known.quit:
            return print("Exiting...", file=sys.stderr)


if __name__ == "__main__":
    main()



import aiohttp
import asyncio
import contextlib
from urllib.parse import urlparse

from .info import API_URL
from .utils import get_running_loop
from .errors import HTTPException


class Threadsafe:
    """
    Threadsafe wrapper for :class:`sans.api.Api` or :class:`sans.api.Dumps` objects.

    This may only be run in a seperate thread from the Api loop.
    This object may be awaited, called, or iterated over.
    Synchronous and asynchronous iteration is supported.

    ================ ============================================================================
    Operation        Description
    ================ ============================================================================
    await x          Make a request to the NS API and returns the root XML element.
    async for y in x Make a request to the NS API and return each shard element as it is parsed.
                     Useful for larger requests.
    x()              Make a request to the NS API and returns the root XML element.
    for y in x       Make a request to the NS API and return each shard element as it is parsed.
                     Useful for larger requests.
    ================ ============================================================================

    Examples
    --------

    Usage::

        sans.run_in_thread()

        darc = await Api(nation="darcania").threadsafe
        async for shard in Api(nation="testlandia").threadsafe:
           print(shard.to_pretty_string())

        tnp = Api(region="the_north_pacific").threadsafe()
        for shard in Api(region="testregionia").threadsafe:
            print(shard.to_pretty_string())
    """

    __slots__ = ("_api",)

    @staticmethod
    def _run_coro_ts(coro):
        from .api import Api

        if not Api._loop or not Api._loop.is_running():
            raise RuntimeError("The API's event loop is not running.")
        if Api._loop == get_running_loop():
            raise RuntimeError(
                "Threaded responses cannot be run in the same loop as the API's loop."
            )
        return asyncio.run_coroutine_threadsafe(coro, Api.loop)

    @staticmethod
    async def _wrapper(awaitable):
        return await awaitable

    def __init__(self, api):
        self._api = api

    def __await__(self):
        return asyncio.wrap_future(
            self._run_coro_ts(self._wrapper(self._api))
        ).__await__()

    async def __aiter__(self):
        aiter = self._api.__aiter__()
        with contextlib.suppress(StopAsyncIteration):
            while True:
                yield (await asyncio.wrap_future(self._run_coro_ts(aiter.__anext__())))

    def __call__(self):
        return self._run_coro_ts(self._wrapper(self._api)).result()

    def __iter__(self):
        aiter = self._api.__aiter__()
        with contextlib.suppress(StopAsyncIteration):
            while True:
                yield self._run_coro_ts(aiter.__anext__()).result()


class NSResponse(aiohttp.ClientResponse):
    __slots__ = ()

    async def start(self, conn):
        from .api import Api

        if urlparse(str(self.real_url))[: len(API_URL)] != API_URL:
            # don't use the lock
            lock = _NullACM()
        else:
            lock = Api._lock
        async with lock:
            for tries in range(5):
                response = await super().start(conn)
                with contextlib.suppress(KeyError):
                    lock._xrlrs(response.headers["X-ratelimit-requests-seen"])
                if response.status == 429:
                    lock._xra(response.headers["X-Retry-After"])
                    break
                elif response.status >= 500:
                    # keep the lock
                    await asyncio.sleep(1 + tries * 2)
                else:
                    break
            return response

    def raise_for_status(self, *args, **kwargs):
        try:
            return super().raise_for_status(*args, **kwargs)
        except aiohttp.ClientResponseError as cre:
            raise HTTPException(cre) from cre


class _NullACM:
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass

    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, name):
        return self



import aiohttp
import asyncio
import contextlib
import functools
import sys
import zlib
from collections.abc import Iterable, Mapping
from datetime import date as Date
from lxml import etree, objectify
from types import MappingProxyType
from typing import (
    Any as _Any,
    AsyncGenerator as _AsyncGenerator,
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)
from urllib.parse import parse_qs, unquote_plus, urlencode, urlparse, urlunparse

from .errors import BadRequest
from .info import API_URL, __version__
from .objects import Threadsafe, NSResponse
from .utils import get_running_loop
from ._lock import ResetLock


__all__ = ["Api", "Dumps", "Archives"]


AGENT_FMT = f"{{}} Python/{sys.version_info[0]}.{sys.version_info[1]} aiohttp/{aiohttp.__version__} sans/{__version__}"
API_VERSION = "10"
PINS: dict = {}


# Set of keys that should be added to rather than overwritten
class _Adds:
    @staticmethod
    def q(x: str, y: str):
        return " ".join((x, y))

    scale, mode, filter, tags = q, q, q, q

    @staticmethod
    def view(x, y):
        xs, ys = x.split("."), y.split(".")
        if len(xs) != 1:
            raise ValueError()
        if len(ys) != 1:
            raise ValueError()
        if xs[0] != ys[0] or len(xs) != 1 or len(ys) != 1:
            # overwrite
            return y
        return "{}.{},{}".format(xs[0], xs[1], ys[1])


def _normalize_dicts(*dicts: _Mapping[str, _Iterable[str]]):
    final: dict = {}
    for d in dicts:
        for k, v in d.items():
            if not all((k, v)):
                continue
            if not isinstance(v, str) and isinstance(v, Iterable):
                v = " ".join(map(str, v))
            v = unquote_plus(str(v)).strip().strip("_")
            if not v:
                continue
            if k in final:
                with contextlib.suppress(AttributeError, TypeError):
                    final[k] = getattr(_Adds, k)(final[k], v)
                    continue
            final[k] = v
    final.setdefault("v", API_VERSION)
    return final


class ApiMeta(type):
    """
    The :class:`Api`'s metaclass.

    Lazily loads and controls access and settings to various global states,
    accessible by attribute access from the :class:`Api` class.
    """

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls._agent = None
        cls.__lock = None
        cls._loop = None
        cls._session = None

    @property
    def agent(cls) -> _Optional[str]:
        """The API wrapper's set user agent."""
        return cls._agent

    @agent.setter
    def agent(cls, value: str) -> None:
        if value:
            cls._agent = AGENT_FMT.format(value)

    @property
    def loop(cls) -> _Optional[asyncio.AbstractEventLoop]:
        """The API wrapper's event loop.

        Cannot be changed once it has been set or referenced."""
        if not cls._loop:
            cls._loop = get_running_loop()
        return cls._loop

    @loop.setter
    def loop(cls, loop: asyncio.AbstractEventLoop) -> None:
        if cls._loop and cls._loop != loop:
            raise asyncio.InvalidStateError("Cannot change loop when it's already set!")
        cls._loop = loop

    @property
    def session(cls) -> aiohttp.ClientSession:
        """The API wrapper's HTTP client session."""
        if not cls._session or cls._session.closed:
            loop = cls.loop
            if not loop:
                cls.loop = asyncio.get_event_loop()
            cls._session = aiohttp.ClientSession(loop=loop, response_class=NSResponse)
        return cls._session

    @property
    def _lock(cls) -> ResetLock:
        if not cls.__lock:
            loop = cls.loop
            if not loop:
                cls.loop = asyncio.get_event_loop()
            cls.__lock = ResetLock(loop=loop)
        return cls.__lock

    @property
    def locked(cls) -> _Optional[bool]:
        """
        Whether or not the API's lock is currently locked.

        As every request acquires the lock, use :attr:`xra` to check if the rate limit is saturated instead.
        """
        if cls.__lock:
            return cls.__lock.locked()
        return None

    @property
    def xra(cls) -> _Optional[float]:
        """
        If the rate limit is currently saturated,
        returns when another API call can be made in the form of a Unix timestamp.
        Otherwise, returns `None`.
        """
        if cls.__lock:
            return cls.__lock.xra
        return None


class Api(metaclass=ApiMeta):
    r"""
    Construct and send an NS API request.
    NS API docs can be found here: https://www.nationstates.net/pages/api.html

    This is a low-level API wrapper. Some attempts will be made to prevent bad requests,
    but it will not check shards against a verified list.
    Authentication may be provided for private nation shards.
    X-Pin headers are be stored internally and globally for ease of use.

    Api objects may be awaited or asynchronously iterated.
    To perform operations from another thread, use the :attr:`threadsafe` property.
    The Api object itself supports all :class:`collections.abc.Mapping` methods.

    ================ ============================================================================
    Operation        Description
    ================ ============================================================================
    await x          Make a request to the NS API and returns the root XML element.
    async for y in x Make a request to the NS API and return each shard element as it is parsed.
                     Useful for larger requests.
    x + y            Combines two :class:`Api` objects together into a new one.
                     Shard keywords that can't be combined will be overwritten with y's data.
    bool(x)          Naïvely check if this :class:`Api` object would result in a 400 Bad Request.
                     Truthy :class:`Api` objects may still result in a 400 Bad Request.
                     Use `len(x)` to check for containment.
    str(x)           Return the URL this :class:`Api` object will make a request to.
    Other            All other :class:`collections.abc.Mapping` methods, except x == y,
                     are also supported.
    ================ ============================================================================

    Parameters
    ----------
    \*shards: str
        Shards to request from the API.
    password: str
        X-Password authentication for private nation shards.
    autologin: str
        X-Autologin authentication for private nation shards.
    \*\*parameters: str
        Query parameters to append to the request, e.g. nation, scale.

    Examples
    --------

    Usage::

        darc = await Api(nation="darcania")
        async for shard in Api(nation="testlandia"):
           print(pretty_string(shard))

        tnp = Api(region="the_north_pacific").threadsafe()
        for shard in Api(region="testregionia").threadsafe:
            print(pretty_string(shard))
    """

    __slots__ = ("__proxy", "_password", "_str", "_hash", "_last_response")

    def __new__(
        cls,
        *shards: _Union[str, _Iterable[str]],
        password: _Optional[str] = None,
        autologin: _Optional[str] = None,
        **parameters: str,
    ):
        if len(shards) == 1 and not parameters:
            if isinstance(shards[0], cls):
                return shards[0]
            with contextlib.suppress(Exception):
                return cls.from_url(shards[0])
        return super().__new__(cls)

    def __init__(
        self,
        *shards: _Union[str, _Iterable[str]],
        password: _Optional[str] = None,
        autologin: _Optional[str] = None,
        **parameters: str,
    ):
        has_nation = "nation" in parameters
        dicts = [parameters] if parameters else []
        for shard in filter(bool, shards):
            if isinstance(shard, Mapping):
                dicts.append(shard)
                if not has_nation and "nation" in shard:
                    has_nation = True
            else:
                dicts.append({"q": shard})
        if not has_nation and (password or autologin):
            raise ValueError("Authentication may only be used with the Nation API.")
        self.__proxy = MappingProxyType(_normalize_dicts(*dicts))
        self._password = password
        self._last_response = None
        self._str = None
        self._hash = None

    async def __await(self):
        async for element in self.__aiter__(clear=False):
            pass
        return element

    def __await__(self):
        return self.__await().__await__()

    async def __aiter__(
        self, *, clear: bool = True
    ) -> _AsyncGenerator[objectify.ObjectifiedElement, None]:
        if not Api.agent:
            raise RuntimeError("The API's user agent is not yet set.")
        if "a" in self and self["a"].lower() == "sendtg":
            raise RuntimeError("This API wrapper does not support API telegrams.")
        if not self:
            # Preempt the request to conserve ratelimit
            raise BadRequest()
        url = str(self)

        headers = {"User-Agent": Api.agent}
        if self._password:
            headers["X-Password"] = self._password
        autologin = self.autologin
        if autologin:
            headers["X-Autologin"] = autologin
        if self.get("nation") in PINS:
            headers["X-Pin"] = PINS[self["nation"]]

        async with Api.session.request("GET", url, headers=headers) as response:
            self._last_response = response
            if "X-Autologin" in response.headers:
                self._password = None
            if "X-Pin" in response.headers:
                PINS[self["nation"]] = response.headers["X-Pin"]
            response.raise_for_status()

            encoding = (
                response.headers["Content-Type"].split("charset=")[1].split(",")[0]
            )
            with contextlib.suppress(etree.XMLSyntaxError), contextlib.closing(
                etree.XMLPullParser(["end"], base_url=url, remove_blank_text=True)
            ) as parser:
                parser.set_element_class_lookup(objectify.ObjectifyElementClassLookup())
                events = parser.read_events()

                async for data, _ in response.content.iter_chunks():
                    parser.feed(data.decode(encoding))
                    for _, element in events:
                        if clear and (
                            element.getparent() is None
                            or element.getparent().getparent() is not None
                        ):
                            continue
                        yield element
                        if clear:
                            element.clear(keep_tail=True)

    def __add__(self, other: _Any) -> "Api":
        if isinstance(other, str):
            with contextlib.suppress(Exception):
                other = type(self).from_url(other)
        with contextlib.suppress(Exception):
            return type(self)(self, other)
        return NotImplemented

    def __bool__(self):
        if any(a in self for a in ("nation", "region")):
            return True
        if "a" in self:
            if self["a"] == "verify" and all(a in self for a in ("nation", "checksum")):
                return True
            if self["a"] == "sendtg" and all(
                a in self for a in ("client", "tgid", "key", "to")
            ):
                return True
            return False
        return "q" in self

    def __contains__(self, key: str) -> bool:
        return key in self.__proxy

    def __dir__(self):
        return set(super().__dir__()).union(
            dir(self.__proxy), (a for a in dir(type(self)) if not hasattr(type, a))
        )

    __eq__ = None

    def __getattr__(self, name: str):
        with contextlib.suppress(AttributeError):
            return getattr(self.__proxy, name)
        raise AttributeError(f"{type(self).__name__!r} has no attribute {name!r}")

    def __getitem__(self, key):
        return self.__proxy[str(key).lower()]

    def __hash__(self):
        if self._hash is not None:
            return self._hash
        params = sorted(
            (k, v if isinstance(v, str) else " ".join(sorted(v)))
            for k, v in self.items()
        )
        self._hash = hash(tuple(params))
        return self._hash

    def __iter__(self):
        return iter(self.__proxy)

    def __len__(self):
        return len(self.__proxy)

    def __repr__(self) -> str:
        return "{}.{}({})".format(
            type(self).__module__,
            type(self).__name__,
            ", ".join("{}={!r}".format(*t) for t in self.__proxy.items()),
        )

    def __str__(self) -> str:
        if self._str is not None:
            return self._str
        params = [
            (k, v if isinstance(v, str) else "+".join(v)) for k, v in self.items()
        ]
        self._str = urlunparse((*API_URL, None, urlencode(params, safe="+"), None))
        return self._str

    @property
    def autologin(self) -> _Optional[str]:
        """
        If a private nation shard was properly requested and returned,
        this property may be used to get the "X-Autologin" token.
        """
        if self._last_response:
            return self._last_response.headers.get("X-Autologin")
        return None

    @property
    def last_headers(self) -> _Optional[_Mapping[str, str]]:
        """
        Returns the headers returned from the last request this API object sent.
        """
        if self._last_response:
            return self._last_response.headers
        return None

    @property
    def last_response(self) -> _Optional[NSResponse]:
        """
        Returns the response object from the last request this API object sent.
        """
        return self._last_response

    @property
    def threadsafe(self) -> Threadsafe:
        """
        Returns a threadsafe wrapper around this object.

        The returned wrapper may be called, awaited, or iterated over.
        Both standard and async iteration are supported.
        """
        return Threadsafe(self)

    @classmethod
    def from_url(
        cls, url: str, *shards: _Iterable[str], **parameters: _Iterable[str]
    ) -> "Api":
        """
        Constructs an Api object from a provided URL.

        The Api object may be further modified with shards and parameters,
        as per the :class:`Api` constructor.
        """
        parsed_url = urlparse(str(url))
        url = parsed_url[: len(API_URL)]
        if any(url) and url != API_URL:
            raise ValueError("URL must be solely query parameters or an API url")
        return cls(*shards, parse_qs(parsed_url.query), parameters)


class _DumpsType:
    __slots__ = "name"

    def __init__(self, name):
        self.name = name

    @functools.lru_cache(maxsize=2)
    def __call__(self, *date: _Union[Date, int]):
        return Dumps(self, *date)

    def __aiter__(self):
        return self().__aiter__()

    def __getattr__(self, attr):
        return getattr(self(), attr)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.name == other.name
        return NotImplemented


class Dumps:
    """
    Request the daily data dumps.
    Dumps docs can be found here: https://www.nationstates.net/pages/api.html#dumps

    Dumps objects support asynchronous iteration, which returns each nation's / region's XML.
    To perform operations from another thread, use the :attr:`threadsafe` property.
    """

    __slots__ = ("url", "_category", "_last_response")

    NATIONS: _ClassVar[_DumpsType] = _DumpsType("nations")
    NATION = NATIONS
    REGIONS: _ClassVar[_DumpsType] = _DumpsType("regions")
    REGION = REGIONS

    def __init__(self, category: _DumpsType, *date: _Union[Date, int]):
        self._category = category
        self._last_response = None
        if not date:
            self.url = urlunparse(
                (*API_URL[:2], "/pages/{name}.xml.gz", None, None, None)
            ).format(name=category.name)
        else:
            if len(date) == 1 and isinstance(date, Date):
                date = date[0]
            else:
                date = Date(*date)
            self.url = urlunparse(
                (*API_URL[:2], "/archive/{name}/{date}-{name}-xml.gz", None, None, None)
            ).format(name=category.name, date=date.isoformat())

    async def __aiter__(self) -> _AsyncGenerator[objectify.ObjectifiedElement, None]:
        if not Api.agent:
            raise RuntimeError("The API's user agent is not yet set.")

        url = self.url
        tag = self._category.name.upper().rstrip("S")
        dobj = zlib.decompressobj(16 + zlib.MAX_WBITS)

        async with Api.session.request(
            "GET", url, headers={"User-Agent": Api.agent}
        ) as response:
            self._last_response = response
            response.raise_for_status()

            with contextlib.suppress(etree.XMLSyntaxError), contextlib.closing(
                etree.XMLPullParser(
                    ["end"], base_url=url, remove_blank_text=True, tag=tag
                )
            ) as parser:
                parser.set_element_class_lookup(objectify.ObjectifyElementClassLookup())
                events = parser.read_events()

                async for data, _ in response.content.iter_chunks():
                    parser.feed(dobj.decompress(data))
                    for _, element in events:
                        yield element
                        element.clear(keep_tail=True)

    @property
    def last_headers(self) -> _Optional[_Mapping[str, str]]:
        """
        Returns the headers returned from the last request this API object sent.
        """
        if self._last_response:
            return self._last_response.headers
        return None

    @property
    def last_response(self) -> _Optional[NSResponse]:
        """
        Returns the response object from the last request this API object sent.
        """
        return self._last_response

    @property
    def threadsafe(self) -> Threadsafe:
        """
        Returns a threadsafe wrapper around this object.

        The returned wrapper may be iterated over.
        Both standard and async iteration are supported.
        """
        return Threadsafe(self)


Archives = Dumps



import aiohttp
import asyncio
import collections
import time


RATE = collections.namedtuple(
    "RateLimit", ("requests", "block", "rpad", "bpad", "retry")
)(50, 30, 2, 0.1, 900)


class ResetLock(asyncio.Lock):
    __slots__ = "__xrlrs", "__xra", "__deferred"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__xrlrs = 0
        self.__xra = None
        self.__deferred = False

    def __aexit__(self, exc_type, exc, tb):
        if isinstance(exc, aiohttp.ClientResponseError):
            if exc.status == 429:
                self._xra(exc.headers["X-Retry-After"])
            elif "X-ratelimit-requests-seen" in exc.headers:
                self._xrlrs(exc.headers["X-ratelimit-requests-seen"])

        # returns a coroutine
        return super().__aexit__(exc_type, exc, tb)

    # Prevent deprecated lock usage
    __await__, __enter__, __iter__ = None, None, None

    def _defer(self):
        self.__deferred = True
        self._loop.call_later(self.__xra - time.time(), self._release)

    def release(self):
        if not self.__deferred:
            super().release()

    def _release(self):
        self.__deferred = False
        super().release()

    def _xra(self, xra: int):
        xra = int(xra)
        self.__xrlrs = 0
        self.__xra = time.time() + xra + RATE.bpad
        self._defer()

    def _xrlrs(self, xrlrs: int):
        now = time.time()
        xrlrs = int(xrlrs)
        if self.__xra is None or xrlrs < self.__xrlrs or self.__xra <= now:
            self.__xra = now + RATE.block + RATE.bpad
        self.__xrlrs = xrlrs
        if xrlrs >= RATE.requests - RATE.rpad:
            self._defer()

    @property
    def xra(self):
        if self.__deferred:
            return self.__xra
        return None

    @property
    def xrlrs(self):
        return self.__xrlrs
